/*
 *
 * This file is part of the Percentile Noise Estimation algorithm.
 *
 * Copyright(c) 2011 Miguel Colom.
 * miguel.colom@cmla.ens-cachan.fr
 *
 * This file may be licensed under the terms of of the
 * GNU General Public License Version 2 (the ``GPL'').
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the GPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the GPL along with this
 * program. If not, go to http://www.gnu.org/licenses/gpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <cmath>
#include <vector>
#include <omp.h>
#include <fstream>
//
#include "algo.h"
#include "curve_filter.h"
#include "CFFT.h"
#include "filters.h"
//
#include "framework/operations.cpp"
#include "framework/CFramework.h"
#include "framework/CImage.h"
#include "framework/libparser.h"
#include "framework/CHistogram.cpp"
#include "framework/CInspector.h"

#define PIXEL(u, N, x, y) (u[(N)*(y) + (x)])

//! Compares a set of pixels
/*!
  \param *data Array of pixels
  \return True if the pixels have the same value
*/
inline bool equals(float *data, int N) {
  for (int i = 0; i < N - 1; i++)
    if (abs(data[i] - data[i+1]) > 1E-3)
      return false;  
  return true;
}

//! Compares the neighborhood of a pixel
/*!
  \param *u Image
  \param x Horizontal coordinate of the pixel
  \param y Vertical coordinate of the pixel
  \param Nx Width of the input image
  \return True if the pixel doesn't belong to a 4-pixels block
*/
inline bool good_pixel(float *u, int x, int y, int Nx) {
  float data[4] = {PIXEL(u, Nx, x, y),
                   PIXEL(u, Nx, x+1, y),
                   PIXEL(u, Nx, x, y+1),
                   PIXEL(u, Nx, x+1, y+1)};
  return !equals(data, 4);
}

//! Initializes the equals pixels mask
/*!
  \param *mask Pixel mask
  \param Nx Image width
  \param Ny Image height
  \param num_channels Number of channels
*/
void init_pixel_mask(float *mask, int Nx, int Ny, int num_channels) {
  for (int q = 0; q < num_channels*Nx*Ny; q++)
    mask[q] = 0;
}

//! Initializes the equals pixels mask
/*!
  \param *mask Pixel masks of all channels
  \param *mask_all Output pixel mask
  \param Nx Image width
  \param Ny Image height
  \param num_channels Number of channels
*/
void consolide_pixel_mask(float *mask, float *mask_all, int Nx, int Ny, int num_channels) {
  for (int q = 0; q < Nx*Ny; q++) {    
    int y = q / Nx;
    int x = q - y*Nx;
    
    int sum = 0;
    for (int ch = 0; ch < num_channels; ch++)
      sum += mask[Nx*Ny*ch + Nx*y + x];
    //mask_all[Nx*y + x] = (sum == num_channels ? 1 : 0); // same pixels in all channels
    mask_all[Nx*y + x] = (sum > 0 ? 1 : 0); // same pixels in any channel
  }
}

//! Add bad pixels to the pixel mask
/*!
  \param *mask Pixel mask
  \param Nx Image width
  \param Ny Image height
  \param w Block side
  \param ch Channel
*/
void add_pixel_mask(float *u, float *mask, int Nx, int Ny, int w, int ch) {  
  for (int q = 0; q < Nx*Ny; q++) {    
    int y = q / Nx;
    int x = q - y*Nx;
    
    if (x < Nx-w+1 && y < Ny-w+1) {
      if (!good_pixel(u, x, y, Nx)) {
        mask[Nx*Ny*ch + Nx*y + x] = 1;
      }
    }
    else
      mask[Nx*Ny*ch + Nx*y + x] = 1;      
  }
}


//! Computes the variances in the filtered images and the means in the original
//! using the wxw blocks.
/*!
  \param *variances Array to store the variances into.
  \param *means Array to store the means into.
  \param *u Original image.
  \param *uf Filtered image.
  \param Nx Width of the image
  \param Ny Height of the image
  \param w Block side
  \param *mask Mask of valid pixels
  \return Number of blocks used
*/
int compute_variances_and_means(float *variances, float *means,
                                   float *u, float *uf,
                                   int Nx, int Ny, int w, float *mask) {
  int w2 = w*w;
  
  int num_x = Nx - w + 1;
  int num_y = Ny - w + 1;

  typedef float  sat_src_type;
  typedef double sat_dst_type;
  
  sat_dst_type *sat_u;
    
  // SAT buffers
  sat_dst_type *sat_uf;  
  sat_src_type *uf2;
  sat_dst_type *sat_uf2;
  
  // Turn off SAT optimization if the image is too big,
  // to avoid running out of memory
  bool turn_off_sats = (Nx*Ny > 6000*6000);
  
  if (turn_off_sats) {
    sat_uf = NULL;  
    uf2 = NULL;
    sat_uf2 = NULL;
  }
  else {
    // Create SAT for u and uf
    sat_u = new sat_dst_type[Nx*Ny];
    sat_uf = new sat_dst_type[Nx*Ny];

    // Create SAT for uf2
    uf2 = new sat_src_type[Nx*Ny];
    for (int i = 0; i < Nx*Ny; i++)
      uf2[i] = uf[i]*uf[i];
    //  
    sat_uf2 = new sat_dst_type[Nx*Ny];

    // Compute SATs
    summed_area_table<sat_src_type, sat_dst_type>(uf2, sat_uf2, Nx, Ny);
    delete[] uf2; uf2 = NULL; // Free memory
    //  
    summed_area_table<sat_src_type, sat_dst_type>(u, sat_u, Nx, Ny);
    summed_area_table<sat_src_type, sat_dst_type>(uf, sat_uf, Nx, Ny);
  }

  int count = 0;
  
  for (int q = 0; q < num_x*num_y; q++) {   
    int y = q / num_x;
    int x = q - y*num_x;
    
    if (mask != NULL && mask[Nx*y+x] == 1) {
      //printf("Discarded block at (%d, %d)\n", x, y);
      continue;
    }

    sat_dst_type s = 0;
    sat_dst_type se2 = 0;
    sat_dst_type mean = 0;

    // Compute mean in u
    if (x > 0 && y > 0 && (!turn_off_sats)) {
      mean = PIXEL(sat_u, Nx, x-1, y-1) + PIXEL(sat_u, Nx, x+w-1, y+w-1) -
              PIXEL(sat_u, Nx, x+w-1, y-1) - PIXEL(sat_u, Nx, x-1, y+w-1);
      //printf("%f\n", mean);
    }
    else {
      for (int j = 0; j < w; j++) {
          for (int i = 0; i < w; i++) {
            float pixel_u = PIXEL(u, Nx, x+i, y+j);
            mean += pixel_u;
          }
      }
    }

    // Compute sum and sum of squares in uf
    if (x > 0 && y > 0 && (!turn_off_sats)) {
      s = PIXEL(sat_uf, Nx, x-1, y-1) + PIXEL(sat_uf, Nx, x+w-1, y+w-1) -
          PIXEL(sat_uf, Nx, x+w-1, y-1) - PIXEL(sat_uf, Nx, x-1, y+w-1);
      se2 = PIXEL(sat_uf2, Nx, x-1, y-1) + PIXEL(sat_uf2, Nx, x+w-1, y+w-1) -
            PIXEL(sat_uf2, Nx, x+w-1, y-1) - PIXEL(sat_uf2, Nx, x-1, y+w-1);
    }
    else {
      for (int j = 0; j < w; j++) {
          for (int i = 0; i < w; i++) {
              float pixel_uf = PIXEL(uf, Nx, x+i, y+j);
              s += pixel_uf;
              se2 += pow(pixel_uf, 2);
          }
      }            
    }

    // Get variance and normalize mean
    float var = se2/(w2) - pow((s/w2), 2);
    if (var < 0) var = 0;
    var *= w2 / (w2 - 1); // Unbiased variance estimator
    //
    mean /=  w2;
    //printf("[%d, %d]: %f, %f\n", x, y, mean, var);
    
    // Store variances and means
    variances[count] = var;
    means[count] = mean;
    count++;
  } // for i

  // Free memory
  if (!turn_off_sats) {
    delete[] sat_u;
    delete[] sat_uf;
    delete[] sat_uf2;
  }

  return count;
}

//! Computes the correction factor to compensate the percentile in pure noise.
/*!
  \param per Percentile
  \param w Block side
  \param op Filter operator
  \return Correction factor
*/
float read_correction_factor(float per, int w, int op, 
                             char *per_corrections_dir) {
  char filename[1000];
  sprintf(filename,
          "%s/per_corr_w%d_p%.4f_f%d.txt",
          per_corrections_dir, w, per, op); 
  
  ifstream infile(filename);
  if (!infile.is_open()) {
      PRINT_ERROR("get_correction_factor failed to open file: %s\n", filename);
      exit(-1);
  }
  
  string line;
  infile >> line;

  float correction = atof(line.c_str());
  
  infile.close();
  return correction;
}

//! Return the maximum number of elements in all the histogram bins
/*!
  \param *histo Histogram
  \return Maximum number of bins
*/
int get_max_elements_bins(CHistogram<float> *histo) {
    int max = 0;
    for (int i = 0; i < histo->get_num_bins(); i++) {
        if (histo->get_num_elements_bin(i) > max)
            max = histo->get_num_elements_bin(i);
    }
    return max;
}

//! Corrects a tilde sigma
/*!
  \param apply_corrections Correction type (0: none, 1: percentile, 2: learning)
  \param tilde_sigma Corrected noise estimation. Used only when correcting with learning
  \param hat_sigma2 Direct noise variance estimation
  \param p Percentile
  \param w Block side
  \param filter_type Prefilter operator
  \param *per_corrections_dir Directory containing the percentile corrections
  \return Corrected noise estimation
*/
float correct_sigma(int apply_corrections, float tilde_sigma, float hat_sigma2,
                    float p, int w, int filter_type, char *per_corrections_dir) {
  switch (apply_corrections) {
      case 0: // No correction
          tilde_sigma = sqrt(hat_sigma2);
          break;
      case 1: { // Percentile correction
          float k = read_correction_factor(p, w,
                                           filter_type,
                                           per_corrections_dir);
          tilde_sigma = sqrt(hat_sigma2) * k;
          break;
      }
  }
  if (tilde_sigma < 0) tilde_sigma = 0;
  return tilde_sigma;  
}

//! Computes the tilde_sigma and means of the data within the bin
/*!
  \param *mean Returned mean
  \param *tilde_sigma Returned tilde_sigma
  \param *bin_variances Variances inside the bin
  \param *bin_means Means inside the bin
  \param num_elements_bin Number of elements in the bin
  \param *indices Sorting indices for the variances in the bin
  \param p Percentile
  \param w Block side
  \param apply_corrections Correction type (0: none, 1: percentile, 2: learning)
  \param filter_type Pre-filter operator type
  \return Corrected noise estimation
*/
void get_tilde_sigma_and_mean(float *mean, float *tilde_sigma,
                              float *bin_variances, float *bin_means,
                              int num_elements_bin, int *indices, float p,
                              int w, int apply_corrections, int filter_type,
                              char *per_corrections_dir) {
        // Apply percentile to the distribution of the bin
        int percentile_pos = round(num_elements_bin * p);
        float hat_sigma2 = num_elements_bin > 0 ?
                           bin_variances[indices[percentile_pos]] : 0;

        *mean = num_elements_bin > 0 ?
                bin_means[indices[percentile_pos]] : 0;

        // Correct hat_sigma2 and get tilde_sigma
        *tilde_sigma = correct_sigma(apply_corrections, *tilde_sigma,
                                     hat_sigma2, p, w, filter_type,
                                     per_corrections_dir);
}

//! Computes the standard deviation and mean of the data
/*!
  \param *std Returned standard deviation
  \param *mean Returned mean
  \param *data Data array
  \param N Number of elements within the data array
*/
void get_std_mean(float *std, float *mean, float *data, int N) {
  // Get mean
  *mean = 0;
  for (int i = 0; i < N; i++)
    *mean += data[i];
  *mean /= N;
  
  // Get variance
  float v = 0;
  for (int i = 0; i < N; i++)
    v += pow(data[i] - *mean, 2);
  v /= N;
  
  *std = sqrt(v);
}

//! Pads the input image by adding (undefined) pixels on the
//! right and bottom.
/*!
  \param *image_from Input image
  \param new_Nx New width
  \param new_Ny New height
  \return Padded image
*/
CImage *change_image_size(CImage *image_from,
                         int new_Nx, int new_Ny) {
  int num_channels = image_from->get_num_channels();

  CImage *image_to = new CImage(new_Nx, new_Ny,
                                image_from->get_bits_per_channel(),
                                num_channels);

  int Nx = image_from->get_width();
  for (int ch = 0; ch < num_channels; ch++) {
    float *from = image_from->get_channel(ch);
    float *to = image_to->get_channel(ch);
    //
    for (int y = 0; y < new_Ny; y++)
      for (int x = 0; x < new_Nx; x++)
        to[new_Nx*y+x] = from[Nx*y+x];
  }
  
  delete image_from;
  return image_to;
}


/*void save_image_debug(float *img, int Nx, int Ny,
                        const char *filename) {
  CImage debug_ima(Nx, Ny, 8, 1);
  float *channel_debug = debug_ima.get_channel(0);
  for (int i = 0; i < Nx*Ny; i++)
    channel_debug[i] = img[i];
  debug_ima.save((char*)filename);
}*/

//! Crops a boundary frame from the given image
/*!
  \param *u Input image
  \param *v Output image
  \param u_Nx Input image width
  \param u_Ny Input image height
  \param border Number of columns or rows that will be cropped from each of the four boundaries of the image
  \return Cropped image
*/
void crop_borders(float *u, float *v, int u_Nx, int u_Ny, int border) {
  int cols_v = u_Nx - 2*border;
  //
  for (int y = border; y < u_Ny-border; y++)
    for (int x = border; x < u_Nx-border; x++)
      v[cols_v*(y-border) + x-border] = u[u_Nx*y + x];
}

//! Creates a vector of helper CFFT objects
/*!
  \param &vCFFT Vector to store the CFFT objets into
  \param Nx Input image width
  \param Ny Input image height
  \param L Side of the LxL filter
*/
void create_FFT_objects(vector <CFFT *>&vCFFT,
                          int Nx, int Ny,
                          int L,
                          int num_channels) {  
  for (int ch = 0; ch < num_channels; ch++) {
    CFFT *FFT = new CFFT(Nx+L-1, Ny+L-1);
    vCFFT.push_back(FFT);
  }
}

//! Deletes the helper CFFT objects contained in the given vector
/*!
  \param &vCFFT Vector storing the CFFT objets
*/
void delete_FFT_objects(vector <CFFT *>vCFFT) {
  int num_channels = vCFFT.size();
  for (int ch = 0; ch < num_channels; ch++) {
    CFFT *FFT = vCFFT[ch];
    delete FFT;
  }
}

void copy_buffer(float *orig, float *dest, int N) {
  for (int i = 0; i < N; i++)
    dest[i] = orig[i];
}

int get_max(int *data, int N) {
  int max = data[0];
  for (int i = 1; i < N; i++)
    if (data[i] > max)
      max = data[i];
  return max;
}

void copy(float *dest, float *orig, int N) {
  for (int i = 0; i < N; i++)
    dest[i] = orig[i];
}


//! Noise estimation Percentile algorithm.
/*!
  \param argc Number of arguments of the program
  \param **argv Arguments of the program
*/
void algorithm(int argc, char **argv) {
  vector <OptStruct *> options;
  vector <ParStruct *> parameters;

  OptStruct obins = {"b:", 0, "0", NULL, "Number of bins"};
  options.push_back(&obins);
  OptStruct oD = {"D:", 7, "7", NULL, "Filtering distance"};
  options.push_back(&oD);  
  OptStruct owin   =  {"w:", 21, "21", NULL, "Variance block"};
  options.push_back(&owin);
  OptStruct opercentile = {"p:", 0.005f, "0.005", NULL, "Percentile"};
  options.push_back(&opercentile);
  OptStruct oprefilter = {"f:", 5, "5", NULL, "Prefilter operator"};
  options.push_back(&oprefilter);
  OptStruct ofiltercurve = {"g:", 3, "3", NULL, "Filter curve iterations"};
  options.push_back(&ofiltercurve);
  OptStruct ore = {"r", 0, NULL, NULL, "Flag to remove equal pixels"};
  options.push_back(&ore);
  OptStruct oh = {"h", 0, NULL, NULL, "Flag to remove outliers"};
  options.push_back(&oh);
  OptStruct oapplyCorrections = {"c:", 1, "1", NULL, "Correction type"};
  options.push_back(&oapplyCorrections);
  OptStruct operCorrectionsDir = {"e:", 9, "./per_corrections/",
                               NULL,
                               "Percentile Gaussian corrections directory"};
  options.push_back(&operCorrectionsDir);
  //
  ParStruct pinput = {"input", NULL, "input file"}; parameters.push_back(&pinput);
  //
  if (!parsecmdline("percentile", "Percentile noise estimation algorithm",
          argc, argv, options, parameters)) {
    printf("\n");
    printf("(c) 2012 Miguel Colom. Under license GNU GPL.\n");
    printf("http://mcolom.perso.math.cnrs.fr/\n");
    printf("\n");
    exit(-1); 
  }
  
  // Parallelization config
  omp_set_num_threads(omp_get_num_procs());

  // Load input image
  CImage *input = new CImage();
  input->load((char*)pinput.value);

  // Read parameters
  int bins = atoi(obins.value);
  int D = atoi(oD.value);    
  int w = atoi(owin.value);
  float p = atof(opercentile.value);
  int filter_type = atof(oprefilter.value);
  int apply_corrections = atoi(oapplyCorrections.value);
  int curve_filter_iterations = atoi(ofiltercurve.value);
  char *per_corrections_dir = (char*)operCorrectionsDir.value;

  // Get image properties
  int Nx = input->get_width();
  int Ny = input->get_height();
  int num_channels = input->get_num_channels();

  // Ensure an odd convolution. The parity of the convolution is the
  // parity of the signal, when convolved with an odd filter.
  if (Nx % 2 == 0 || Ny % 2 == 0) {
    if (Nx % 2 == 0) Nx--;
    if (Ny % 2 == 0) Ny--;
    input = change_image_size(input, Nx, Ny);
  }

  // Automatic selection of bins?
  if (bins <= 0) bins = Nx * Ny / 42000;
  if (bins <= 0) bins = 1; // Force at least one bin

  int max_len_variances = Nx * Ny;
  if (max_len_variances <= 0) {
      printf("# Size of the image too small!\n");
      exit(-1);
  }

  // Vectors to store the final means and noise estimations
  float *vmeans = new float[num_channels * bins]; // [ch*bins + bin]
  float *vstds = new float[num_channels * bins];  // [ch*bins + bin]
    
  // Create prefilter matrix
  int L; // Filter support
  float *filter = get_filter(&L, filter_type);  
  
  float *mask_all;
  if (ore.flag) { // remove groups of 4 equal pixels
    float *mask = new float[num_channels*Nx*Ny];
    mask_all = new float[Nx*Ny];
    init_pixel_mask(mask, Nx, Ny, num_channels);

    for (int ch = 0; ch < num_channels; ch++) {
      float *u = input->get_channel(ch);
      add_pixel_mask(u, mask, Nx, Ny, w, ch);
    } 
    consolide_pixel_mask(mask, mask_all, Nx, Ny, num_channels);
    delete[] mask;

    // Cut mask
    float *mask_cut = new float[(Nx-2*L)*(Ny-2*L)];
    crop_borders(mask_all, mask_cut, Nx, Ny, L);
    delete mask_all;
    mask_all = mask_cut;
  }
  else
    mask_all = NULL;

  vector <CFFT *> vCFFT;
  create_FFT_objects(vCFFT, Nx, Ny, L, num_channels);

  // Process each channel
  #ifdef _OPENMP
  #pragma omp parallel for
  #endif
  for (int ch = 0; ch < num_channels; ch++) {
      float *u = input->get_channel(ch);

      float *uf = new float[(Nx+L-1)*(Ny+L-1)]; // u filtered
      CFFT *FFT = vCFFT[ch]; // FFTW isn't reentrant.
      prefilter(u, uf, Nx, Ny, filter, L, FFT);

      float *u_cut = new float[(Nx-L+1)*(Ny-L+1)];

      // Crop (L-1)/2 pixels from the corners
      crop_borders(u, u_cut, Nx, Ny, (L-1)/2);
      // Now u_cut is (Nx-F+1, Ny-F+1)
    
      // Crop L-1 pixels from the corners
      float *tmp = new float[(Nx-L+1)*(Ny-L+1)];
      crop_borders(uf, tmp, Nx+L-1, Ny+L-1, L-1);
      copy_buffer(tmp, uf, (Nx-L+1)*(Ny-L+1));
      delete[] tmp;
      // Now uf is (Nx-F+1, Ny-F+1), the same size as u_cut
      
      float *variances = new float[max_len_variances];
      float *means = new float[max_len_variances];

      int count = compute_variances_and_means(variances, means,
                                              u_cut, uf,
                                              Nx-L+1, Ny-L+1, w, mask_all);
      if (count == 0) {
          printf("# Size of the image too small!\n");
          exit(-1);
      }
      
      // Create histogram
      CHistogram<float> histo = CHistogram<float>(bins, variances,
                                                  means, count,
                                                  true);

      // Process each bin of the histogram
      
      int max_elements_bins = get_max_elements_bins(&histo);
      int *indices_all = new int[max_elements_bins * bins];

      // Compute and store the indices for all bins
      for (int bin = 0; bin < bins; bin++) {
        float *bin_variances = histo.get_data_bin(bin);
        int num_elements_bin = histo.get_num_elements_bin(bin);

        argsort<float>(bin_variances,
                       &indices_all[max_elements_bins * bin],
                       num_elements_bin);
      }

      for (int bin = 0; bin < bins; bin++) {
        float mean;
        float tilde_sigma;
        
        float *bin_variances = histo.get_data_bin(bin);
        float *bin_means = histo.get_datal_bin(bin);

        int num_elements_bin = histo.get_num_elements_bin(bin);
        int *indices = &indices_all[max_elements_bins * bin];
        
        get_tilde_sigma_and_mean(&mean, &tilde_sigma,
                              bin_variances, bin_means,
                              num_elements_bin, indices, p,
                              w, apply_corrections, filter_type,
                              per_corrections_dir);

        // Remove outliers
        if (oh.flag) { // remove outliers inside the bin
          CFramework *fw = CFramework::get_framework();
          for (int iters = 0; iters < 100; iters++) {
            // Remove outliers
            float new_sigma;
            float *without_outliers_vars = fw->create_array(num_elements_bin);
            float *without_outliers_means = fw->create_array(num_elements_bin);

            float bin_std = 0;
            float bin_mean = 0;
            get_std_mean(&bin_std, &bin_mean, bin_variances, num_elements_bin);

            int num_no_outliers = 0;
            for (int i = 0; i < num_elements_bin; i++) {
              if (abs(bin_variances[i] - pow(tilde_sigma,2)) < 2*bin_std) {
                without_outliers_vars[num_no_outliers] = bin_variances[i];
                without_outliers_means[num_no_outliers] = bin_means[i];
                num_no_outliers++;
              }
            }

            for (int i = 0; i < num_no_outliers; i++) {
              bin_variances[i] = without_outliers_vars[i];
              bin_means[i] = without_outliers_means[i];
            }
            num_elements_bin = num_no_outliers;

            argsort<float>(bin_variances,
                  indices,
                  num_elements_bin);
            //
            get_tilde_sigma_and_mean(&mean, &new_sigma,
                                  bin_variances, bin_means,
                                  num_elements_bin, indices, p,
                                  w, apply_corrections, filter_type,
                                  per_corrections_dir);

            if (new_sigma >= tilde_sigma)
              break;
            else
              tilde_sigma = new_sigma;          
          }
        }

        // Store results
        vmeans[ch*bins + bin] = mean;
        vstds[ch*bins + bin] = tilde_sigma;
      } // for bin

      delete[] indices_all;
      delete[] u_cut;
      delete[] uf;
  }
  
  delete_FFT_objects(vCFFT);

  // Filter noise curve
  float *new_std_control = new float[bins * num_channels];
  copy(new_std_control, vstds, num_channels*bins);
  //
  for (int ch = 0; ch < num_channels; ch++)
    for (int filt_iter = 0; filt_iter < curve_filter_iterations; filt_iter++) {
      bool allow_up = (filt_iter < 3);
      filter_curve(&vmeans[ch*bins], &new_std_control[ch*bins],
                   bins,
                   &new_std_control[ch*bins],
                   D, allow_up);
    }
  
  // Print results
  for (int bin = 0; bin < bins; bin++) {
    // Means
    for (int ch = 0; ch < num_channels; ch++)
        printf("%f  ", vmeans[ch*bins+bin]);
        
    // Standard deviations
    for (int ch = 0; ch < num_channels; ch++)
        printf("%f  ", new_std_control[ch*bins+bin]);
    //
    printf("\n");
  }

  // Cleanup
  delete input;
  if (mask_all != NULL) delete[] mask_all;
  delete[] new_std_control;  
  delete[] vmeans;
  delete[] vstds;
}
